@@ -38,7 +38,6 @@
 
 ;; Pacify byte-compiler.
 (defvar directory-sep-char)
-(defvar dired-move-to-filename-regexp)
 
 (defcustom tramp-adb-program "adb"
   "Name of the Android Debug Bridge program."
@@ -134,13 +133,13 @@
     (find-backup-file-name . tramp-handle-find-backup-file-name)
     ;; `find-file-noselect' performed by default handler.
     ;; `get-file-buffer' performed by default handler.
-    (insert-directory . tramp-adb-handle-insert-directory)
+    (insert-directory . tramp-handle-insert-directory)
     (insert-file-contents . tramp-handle-insert-file-contents)
     (load . tramp-handle-load)
     (make-auto-save-file-name . tramp-handle-make-auto-save-file-name)
     (make-directory . tramp-adb-handle-make-directory)
     (make-directory-internal . ignore)
-    (make-symbolic-link . ignore)
+    (make-symbolic-link . tramp-handle-make-symbolic-link)
     (process-file . tramp-adb-handle-process-file)
     (rename-file . tramp-adb-handle-rename-file)
     (set-file-acl . ignore)
@@ -187,7 +186,8 @@ pass to the OPERATION."
 	    result)
 	(tramp-compat-set-process-query-on-exit-flag p nil)
 	(while (eq 'run (process-status p))
-	  (sleep-for 0.1))
+	  (accept-process-output p 0.1))
+	(accept-process-output p 0.1)
 	(goto-char (point-min))
 	(while (search-forward-regexp "^\\(\\S-+\\)[[:space:]]+device$" nil t)
 	  (add-to-list 'result (list nil (match-string 1))))
@@ -228,98 +228,105 @@ pass to the OPERATION."
 ;; code could be shared?
 (defun tramp-adb-handle-file-truename (filename)
   "Like `file-truename' for Tramp files."
-  (with-parsed-tramp-file-name (expand-file-name filename) nil
-    (with-tramp-file-property v localname "file-truename"
-      (let ((result nil))			; result steps in reverse order
-	(tramp-message v 4 "Finding true name for `%s'" filename)
-	(let* ((directory-sep-char ?/)
-	       (steps (tramp-compat-split-string localname "/"))
-	       (localnamedir (tramp-run-real-handler
-			      'file-name-as-directory (list localname)))
-	       (is-dir (string= localname localnamedir))
-	       (thisstep nil)
-	       (numchase 0)
-	       ;; Don't make the following value larger than
-	       ;; necessary.  People expect an error message in a
-	       ;; timely fashion when something is wrong; otherwise
-	       ;; they might think that Emacs is hung.  Of course,
-	       ;; correctness has to come first.
-	       (numchase-limit 20)
-	       symlink-target)
-	  (while (and steps (< numchase numchase-limit))
-	    (setq thisstep (pop steps))
-	    (tramp-message
-	     v 5 "Check %s"
-	     (mapconcat 'identity
-			(append '("") (reverse result) (list thisstep))
-			"/"))
-	    (setq symlink-target
-		  (nth 0 (file-attributes
-			  (tramp-make-tramp-file-name
-			   method user host
-			   (mapconcat 'identity
-				      (append '("")
-					      (reverse result)
-					      (list thisstep))
-				      "/")))))
-	    (cond ((string= "." thisstep)
-		   (tramp-message v 5 "Ignoring step `.'"))
-		  ((string= ".." thisstep)
-		   (tramp-message v 5 "Processing step `..'")
-		   (pop result))
-		  ((stringp symlink-target)
-		   ;; It's a symlink, follow it.
-		   (tramp-message v 5 "Follow symlink to %s" symlink-target)
-		   (setq numchase (1+ numchase))
-		   (when (file-name-absolute-p symlink-target)
-		     (setq result nil))
-		   ;; If the symlink was absolute, we'll get a string
-		   ;; like "/user@host:/some/target"; extract the
-		   ;; "/some/target" part from it.
-		   (when (tramp-tramp-file-p symlink-target)
-		     (unless (tramp-equal-remote filename symlink-target)
-		       (tramp-error
-			v 'file-error
-			"Symlink target `%s' on wrong host" symlink-target))
-		     (setq symlink-target localname))
-		   (setq steps
-			 (append (tramp-compat-split-string
-				  symlink-target "/")
-				 steps)))
-		  (t
-		   ;; It's a file.
-		   (setq result (cons thisstep result)))))
-	  (when (>= numchase numchase-limit)
-	    (tramp-error
-	     v 'file-error
-	     "Maximum number (%d) of symlinks exceeded" numchase-limit))
-	  (setq result (reverse result))
-	  ;; Combine list to form string.
-	  (setq result
-		(if result
-		    (mapconcat 'identity (cons "" result) "/")
-		  "/"))
-	  (when (and is-dir (or (string= "" result)
-				(not (string= (substring result -1) "/"))))
-	    (setq result (concat result "/"))))
+  (format
+   "%s%s"
+   (with-parsed-tramp-file-name (expand-file-name filename) nil
+     (tramp-make-tramp-file-name
+      method user host
+      (with-tramp-file-property v localname "file-truename"
+	(let ((result nil))			; result steps in reverse order
+	  (tramp-message v 4 "Finding true name for `%s'" filename)
+	  (let* ((directory-sep-char ?/)
+		 (steps (tramp-compat-split-string localname "/"))
+		 (localnamedir (tramp-run-real-handler
+				'file-name-as-directory (list localname)))
+		 (is-dir (string= localname localnamedir))
+		 (thisstep nil)
+		 (numchase 0)
+		 ;; Don't make the following value larger than
+		 ;; necessary.  People expect an error message in a
+		 ;; timely fashion when something is wrong; otherwise
+		 ;; they might think that Emacs is hung.  Of course,
+		 ;; correctness has to come first.
+		 (numchase-limit 20)
+		 symlink-target)
+	    (while (and steps (< numchase numchase-limit))
+	      (setq thisstep (pop steps))
+	      (tramp-message
+	       v 5 "Check %s"
+	       (mapconcat 'identity
+			  (append '("") (reverse result) (list thisstep))
+			  "/"))
+	      (setq symlink-target
+		    (nth 0 (file-attributes
+			    (tramp-make-tramp-file-name
+			     method user host
+			     (mapconcat 'identity
+					(append '("")
+						(reverse result)
+						(list thisstep))
+					"/")))))
+	      (cond ((string= "." thisstep)
+		     (tramp-message v 5 "Ignoring step `.'"))
+		    ((string= ".." thisstep)
+		     (tramp-message v 5 "Processing step `..'")
+		     (pop result))
+		    ((stringp symlink-target)
+		     ;; It's a symlink, follow it.
+		     (tramp-message v 5 "Follow symlink to %s" symlink-target)
+		     (setq numchase (1+ numchase))
+		     (when (file-name-absolute-p symlink-target)
+		       (setq result nil))
+		     ;; If the symlink was absolute, we'll get a string
+		     ;; like "/user@host:/some/target"; extract the
+		     ;; "/some/target" part from it.
+		     (when (tramp-tramp-file-p symlink-target)
+		       (unless (tramp-equal-remote filename symlink-target)
+			 (tramp-error
+			  v 'file-error
+			  "Symlink target `%s' on wrong host" symlink-target))
+		       (setq symlink-target localname))
+		     (setq steps
+			   (append (tramp-compat-split-string
+				    symlink-target "/")
+				   steps)))
+		    (t
+		     ;; It's a file.
+		     (setq result (cons thisstep result)))))
+	    (when (>= numchase numchase-limit)
+	      (tramp-error
+	       v 'file-error
+	       "Maximum number (%d) of symlinks exceeded" numchase-limit))
+	    (setq result (reverse result))
+	    ;; Combine list to form string.
+	    (setq result
+		  (if result
+		      (mapconcat 'identity (cons "" result) "/")
+		    "/"))
+	    (when (and is-dir (or (string= "" result)
+				  (not (string= (substring result -1) "/"))))
+	      (setq result (concat result "/"))))
 
-        (tramp-message v 4 "True name of `%s' is `%s'" filename result)
-        (tramp-make-tramp-file-name method user host result)))))
+	  (tramp-message v 4 "True name of `%s' is `%s'" localname result)
+	  result))))
+
+   ;; Preserve trailing "/".
+   (if (string-equal (file-name-nondirectory filename) "") "/" "")))
 
 (defun tramp-adb-handle-file-attributes (filename &optional id-format)
   "Like `file-attributes' for Tramp files."
   (unless id-format (setq id-format 'integer))
-  (ignore-errors
-    (with-parsed-tramp-file-name filename nil
-      (with-tramp-file-property
-	  v localname (format "file-attributes-%s" id-format)
-	(tramp-adb-barf-unless-okay
-	 v (format "%s -d -l %s"
-		   (tramp-adb-get-ls-command v)
-		   (tramp-shell-quote-argument localname)) "")
-	(with-current-buffer (tramp-get-buffer v)
-	  (tramp-adb-sh-fix-ls-output)
-	  (cdar (tramp-do-parse-file-attributes-with-ls v id-format)))))))
+  (with-parsed-tramp-file-name filename nil
+    (with-tramp-file-property
+	v localname (format "file-attributes-%s" id-format)
+      (and
+       (tramp-adb-send-command-and-check
+	v (format "%s -d -l %s"
+		  (tramp-adb-get-ls-command v)
+		  (tramp-shell-quote-argument localname)))
+       (with-current-buffer (tramp-get-buffer v)
+	 (tramp-adb-sh-fix-ls-output)
+	 (cdar (tramp-do-parse-file-attributes-with-ls v id-format)))))))
 
 (defun tramp-do-parse-file-attributes-with-ls (vec &optional id-format)
   "Parse `file-attributes' for Tramp files using the ls(1) command."
@@ -366,11 +373,19 @@ pass to the OPERATION."
       (with-tramp-file-property
 	  v localname (format "directory-files-attributes-%s-%s-%s-%s"
 			      full match id-format nosort)
-	(tramp-adb-barf-unless-okay
-	 v (format "%s -a -l %s"
-		   (tramp-adb-get-ls-command v)
-		   (tramp-shell-quote-argument localname)) "")
 	(with-current-buffer (tramp-get-buffer v)
+	  (when (tramp-adb-send-command-and-check
+		 v (format "%s -a -l %s"
+			   (tramp-adb-get-ls-command v)
+			   (tramp-shell-quote-argument localname)))
+	    ;; We insert also filename/. and filename/.., because "ls" doesn't.
+	    (narrow-to-region (point) (point))
+	    (tramp-adb-send-command
+	     v (format "%s -d -a -l %s %s"
+		       (tramp-adb-get-ls-command v)
+		       (concat (file-name-as-directory localname) ".")
+		       (concat (file-name-as-directory localname) "..")))
+	    (widen))
 	  (tramp-adb-sh-fix-ls-output)
 	  (let ((result (tramp-do-parse-file-attributes-with-ls
 			 v (or id-format 'integer))))
@@ -392,8 +407,7 @@ pass to the OPERATION."
 (defun tramp-adb-get-ls-command (vec)
   (with-tramp-connection-property vec "ls"
     (tramp-message vec 5 "Finding a suitable `ls' command")
-    (if	(zerop (tramp-adb-command-exit-status
-		vec "ls --color=never -al /dev/null"))
+    (if (tramp-adb-send-command-and-check vec "ls --color=never -al /dev/null")
 	;; On CyanogenMod based system BusyBox is used and "ls" output
 	;; coloring is enabled by default.  So we try to disable it
 	;; when possible.
@@ -417,35 +431,6 @@ Convert (\"-al\") to (\"-a\" \"-l\").  Remove arguments like \"--dired\"."
 			   (and (not (string-match "\\(^--\\|^[^-]\\)" s)) s))
 			 switches))))))
 
-(defun tramp-adb-handle-insert-directory
-  (filename switches &optional _wildcard _full-directory-p)
-  "Like `insert-directory' for Tramp files."
-  (when (stringp switches)
-    (setq switches (tramp-adb--gnu-switches-to-ash (split-string switches))))
-  (with-parsed-tramp-file-name (file-truename filename) nil
-    (with-current-buffer (tramp-get-buffer v)
-      (let ((name (tramp-shell-quote-argument (directory-file-name localname)))
-	    (switch-d (member "-d" switches))
-	    (switch-t (member "-t" switches))
-	    (switches (mapconcat 'identity (remove "-t" switches) " ")))
-	(tramp-adb-barf-unless-okay
-	 v (format "%s %s %s" (tramp-adb-get-ls-command v) switches name)
-	 "Cannot insert directory listing: %s" filename)
-	(unless switch-d
-	  ;; We insert also filename/. and filename/.., because "ls" doesn't.
-	  (narrow-to-region (point) (point))
-	  (ignore-errors
-	    (tramp-adb-barf-unless-okay
-	     v (format "%s -d %s %s %s"
-		       (tramp-adb-get-ls-command v)
-		       switches
-		       (concat (file-name-as-directory name) ".")
-		       (concat (file-name-as-directory name) ".."))
-	     "Cannot insert directory listing: %s" filename))
-	  (widen))
-	(tramp-adb-sh-fix-ls-output switch-t)))
-    (insert-buffer-substring (tramp-get-buffer v))))
-
 (defun tramp-adb-sh-fix-ls-output (&optional sort-by-time)
   "Insert dummy 0 in empty size columns.
 Androids \"ls\" command doesn't insert size column for directories:
@@ -489,9 +474,9 @@ Emacs dired can't find files."
 (defun tramp-adb-ls-output-name-less-p (a b)
   "Sort \"ls\" output by name, ascending."
   (let (posa posb)
-    (string-match dired-move-to-filename-regexp a)
+    (string-match directory-listing-before-filename-regexp a)
     (setq posa (match-end 0))
-    (string-match dired-move-to-filename-regexp b)
+    (string-match directory-listing-before-filename-regexp b)
     (setq posb (match-end 0))
     (string-lessp (substring a posa) (substring b posb))))
 
@@ -506,7 +491,8 @@ Emacs dired can't find files."
     (tramp-adb-barf-unless-okay
      v (format "mkdir %s" (tramp-shell-quote-argument localname))
      "Couldn't make directory %s" dir)
-    (tramp-flush-directory-property v (file-name-directory localname))))
+    (tramp-flush-file-property v (file-name-directory localname))
+    (tramp-flush-directory-property v localname)))
 
 (defun tramp-adb-handle-delete-directory (directory &optional recursive)
   "Like `delete-directory' for Tramp files."
@@ -538,20 +524,22 @@ Emacs dired can't find files."
      (with-tramp-file-property v localname "file-name-all-completions"
        (save-match-data
 	 (tramp-adb-send-command
-	  v (format "%s %s"
+	  v (format "%s -a %s"
 		    (tramp-adb-get-ls-command v)
 		    (tramp-shell-quote-argument localname)))
 	 (mapcar
 	  (lambda (f)
-	    (if (file-directory-p f)
+	    (if (file-directory-p (expand-file-name f directory))
 		(file-name-as-directory f)
 	      f))
 	  (with-current-buffer (tramp-get-buffer v)
-	    (delq
-	     nil
-	     (mapcar
-	      (lambda (l) (and (not (string-match  "^[[:space:]]*$" l)) l))
-	      (split-string (buffer-string) "\n"))))))))))
+	    (append
+	     '("." "..")
+	     (delq
+	      nil
+	      (mapcar
+	       (lambda (l) (and (not (string-match  "^[[:space:]]*$" l)) l))
+	       (split-string (buffer-string) "\n")))))))))))
 
 (defun tramp-adb-handle-file-local-copy (filename)
   "Like `file-local-copy' for Tramp files."
@@ -567,7 +555,10 @@ Emacs dired can't find files."
 	  (delete-file tmpfile)
 	  (tramp-error
 	   v 'file-error "Cannot make local copy of file `%s'" filename))
-	(set-file-modes tmpfile (file-modes filename)))
+	(set-file-modes
+	 tmpfile
+	 (logior (or (file-modes filename) 0)
+		 (tramp-compat-octal-to-decimal "0400"))))
       tmpfile)))
 
 (defun tramp-adb-handle-file-writable-p (filename)
@@ -577,9 +568,8 @@ But handle the case, if the \"test\" command is not available."
     (with-tramp-file-property v localname "file-writable-p"
       (if (tramp-adb-find-test-command v)
 	  (if (file-exists-p filename)
-	      (zerop
-	       (tramp-adb-command-exit-status
-		v (format "test -w %s" (tramp-shell-quote-argument localname))))
+	      (tramp-adb-send-command-and-check
+	       v (format "test -w %s" (tramp-shell-quote-argument localname)))
 	    (and
 	     (file-directory-p (file-name-directory filename))
 	     (file-writable-p (file-name-directory filename))))
@@ -599,9 +589,6 @@ But handle the case, if the \"test\" command is not available."
   "Like `write-region' for Tramp files."
   (setq filename (expand-file-name filename))
   (with-parsed-tramp-file-name filename nil
-    (when append
-      (tramp-error
-       v 'file-error "Cannot append to file using Tramp (`%s')" filename))
     (when (and confirm (file-exists-p filename))
       (unless (y-or-n-p (format "File %s exists; overwrite anyway? "
 				filename))
@@ -612,14 +599,20 @@ But handle the case, if the \"test\" command is not available."
     (tramp-flush-file-property v localname)
     (let* ((curbuf (current-buffer))
 	   (tmpfile (tramp-compat-make-temp-file filename)))
+      (when (and append (file-exists-p filename))
+	(copy-file filename tmpfile 'ok)
+	(set-file-modes
+	 tmpfile
+	 (logior (or (file-modes tmpfile) 0)
+		 (tramp-compat-octal-to-decimal "0600"))))
       (tramp-run-real-handler
        'write-region
        (list start end tmpfile append 'no-message lockname confirm))
       (with-tramp-progress-reporter
-	  v 3 (format "Moving tmp file %s to %s" tmpfile filename)
+	  v 3 (format "Moving tmp file `%s' to `%s'" tmpfile filename)
 	(unwind-protect
 	    (when (tramp-adb-execute-adb-command v "push" tmpfile localname)
-	      (tramp-error v 'file-error "Cannot write: `%s' filename"))
+	      (tramp-error v 'file-error "Cannot write: `%s'" filename))
 	  (delete-file tmpfile)))
 
       (when (or (eq visit t) (stringp visit))
@@ -634,9 +627,8 @@ But handle the case, if the \"test\" command is not available."
   "Like `set-file-modes' for Tramp files."
   (with-parsed-tramp-file-name filename nil
     (tramp-flush-file-property v localname)
-    (tramp-adb-barf-unless-okay
-     v (format "chmod %s %s" (tramp-compat-decimal-to-octal mode) localname)
-     "Error while changing file's mode %s" filename)))
+    (tramp-adb-send-command-and-check
+     v (format "chmod %s %s" (tramp-compat-decimal-to-octal mode) localname))))
 
 (defun tramp-adb-handle-set-file-times (filename &optional time)
   "Like `set-file-times' for Tramp files."
@@ -645,8 +637,8 @@ But handle the case, if the \"test\" command is not available."
     (let ((time (if (or (null time) (equal time '(0 0)))
 		    (current-time)
 		  time)))
-      (tramp-adb-command-exit-status
-       ;; use shell arithmetic because of Emacs integer size limit
+      (tramp-adb-send-command-and-check
+       ;; Use shell arithmetic because of Emacs integer size limit.
        v (format "touch -t $(( %d * 65536 + %d )) %s"
 		 (car time) (cadr time)
 		 (tramp-shell-quote-argument localname))))))
@@ -704,32 +696,36 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
   (setq filename (expand-file-name filename)
 	newname (expand-file-name newname))
 
-  (with-parsed-tramp-file-name
-      (if (tramp-tramp-file-p filename) filename newname) nil
-    (with-tramp-progress-reporter
-	v 0 (format "Renaming %s to %s" newname filename)
+  (let ((t1 (tramp-tramp-file-p filename))
+	(t2 (tramp-tramp-file-p newname)))
+    (with-parsed-tramp-file-name (if t1 filename newname) nil
+      (with-tramp-progress-reporter
+	  v 0 (format "Renaming %s to %s" filename newname)
 
-      (if (and (tramp-equal-remote filename newname)
-	       (not (file-directory-p filename)))
-	  (progn
-	    (when (and (not ok-if-already-exists)
-		       (file-exists-p newname))
-	      (tramp-error v 'file-already-exists newname))
-	    ;; We must also flush the cache of the directory, because
-	    ;; `file-attributes' reads the values from there.
-	    (tramp-flush-file-property v (file-name-directory localname))
-	    (tramp-flush-file-property v localname)
-	    ;; Short track.
-	    (tramp-adb-barf-unless-okay
-	     v (format
-		"mv %s %s"
-		(tramp-file-name-handler 'file-remote-p filename 'localname)
-		localname)
-	     "Error renaming %s to %s" filename newname))
+	(if (and t1 t2
+		 (tramp-equal-remote filename newname)
+		 (not (file-directory-p filename)))
+	    (let ((l1 (tramp-file-name-handler
+		       'file-remote-p filename 'localname))
+		  (l2 (tramp-file-name-handler
+		       'file-remote-p newname 'localname)))
+	      (when (and (not ok-if-already-exists)
+			 (file-exists-p newname))
+		(tramp-error v 'file-already-exists newname))
+	      ;; We must also flush the cache of the directory, because
+	      ;; `file-attributes' reads the values from there.
+	      (tramp-flush-file-property v (file-name-directory l1))
+	      (tramp-flush-file-property v l1)
+	      (tramp-flush-file-property v (file-name-directory l2))
+	      (tramp-flush-file-property v l2)
+	      ;; Short track.
+	      (tramp-adb-barf-unless-okay
+	       v (format "mv %s %s" l1 l2)
+	       "Error renaming %s to %s" filename newname))
 
-	;; Rename by copy.
-	(copy-file filename newname ok-if-already-exists t t)
-	(delete-file filename)))))
+	  ;; Rename by copy.
+	  (copy-file filename newname ok-if-already-exists t t)
+	  (delete-file filename))))))
 
 (defun tramp-adb-handle-process-file
   (program &optional infile destination display &rest args)
@@ -1002,7 +998,8 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
   (with-temp-buffer
     (prog1
 	(unless
-	    (zerop (apply 'tramp-call-process tramp-adb-program nil t nil args))
+	    (zerop
+	     (apply 'tramp-call-process vec tramp-adb-program nil t nil args))
 	  (buffer-string))
       (tramp-message vec 6 "%s" (buffer-string)))))
 
@@ -1010,7 +1007,7 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
   "Checks, whether the ash has a builtin \"test\" command.
 This happens for Android >= 4.0."
   (with-tramp-connection-property vec "test"
-    (zerop (tramp-adb-command-exit-status vec "type test"))))
+    (tramp-adb-send-command-and-check vec "type test")))
 
 ;; Connection functions
 
@@ -1033,9 +1030,9 @@ This happens for Android >= 4.0."
       (while (re-search-forward "\r+$" nil t)
 	(replace-match "" nil nil)))))
 
-(defun tramp-adb-command-exit-status
+(defun tramp-adb-send-command-and-check
   (vec command)
-  "Run COMMAND and return its exit status.
+  "Run COMMAND and and check its exit status.
 Sends `echo $?' along with the COMMAND for checking the exit status.  If
 COMMAND is nil, just sends `echo $?'.  Returns the exit status found."
   (tramp-adb-send-command
@@ -1049,14 +1046,14 @@ COMMAND is nil, just sends `echo $?'.  Returns the exit status found."
        vec 'file-error "Couldn't find exit status of `%s'" command))
     (skip-chars-forward "^ ")
     (prog1
-	(read (current-buffer))
+	(zerop (read (current-buffer)))
       (let (buffer-read-only)
 	(delete-region (match-beginning 0) (point-max))))))
 
 (defun tramp-adb-barf-unless-okay (vec command fmt &rest args)
   "Run COMMAND, check exit status, throw error if exit status not okay.
 FMT and ARGS are passed to `error'."
-  (unless (zerop (tramp-adb-command-exit-status vec command))
+  (unless (tramp-adb-send-command-and-check vec command)
     (apply 'tramp-error vec 'file-error fmt args)))
 
 (defun tramp-adb-wait-for-output (proc &optional timeout)
@@ -1099,7 +1096,7 @@ connection if a previous connection has died for some reason."
 	 (p (get-buffer-process buf))
 	 (host (tramp-file-name-host vec))
 	 (user (tramp-file-name-user vec))
-	 (devices (mapcar 'cadr (tramp-adb-parse-device-names nil))))
+	 devices)
 
     ;; Maybe we know already that "su" is not supported.  We cannot
     ;; use a connection property, because we have not checked yet
@@ -1111,6 +1108,7 @@ connection if a previous connection has died for some reason."
 	(and p (processp p) (memq (process-status p) '(run open)))
       (save-match-data
 	(when (and p (processp p)) (delete-process p))
+	(setq devices (mapcar 'cadr (tramp-adb-parse-device-names nil)))
 	(if (not devices)
 	    (tramp-error vec 'file-error "No device connected"))
 	(if (and (> (length host) 0) (not (member host devices)))
@@ -1165,7 +1163,7 @@ connection if a previous connection has died for some reason."
 	    ;; Change user if indicated.
 	    (when user
 	      (tramp-adb-send-command vec (format "su %s" user))
-	      (unless (zerop (tramp-adb-command-exit-status vec nil))
+	      (unless (tramp-adb-send-command-and-check vec nil)
 		(delete-process p)
 		(tramp-set-file-property vec "" "su-command-p" nil)
 		(tramp-error
@@ -1183,5 +1181,9 @@ connection if a previous connection has died for some reason."
 		(read (current-buffer)))
 	      ":" 'omit-nulls))))))))
 
+(add-hook 'tramp-unload-hook
+	  (lambda ()
+	    (unload-feature 'tramp-adb 'force)))
+
 (provide 'tramp-adb)
 ;;; tramp-adb.el ends here
