@@ -447,8 +447,7 @@ pass to the OPERATION."
 			       (expand-file-name
 				tramp-temp-name-prefix
 				(tramp-compat-temporary-file-directory))))
-		   (args      (list tramp-smb-program
-				    (concat "//" real-host "/" share) "-E")))
+		   (args      (list (concat "//" real-host "/" share) "-E")))
 
 	      (if (not (zerop (length real-user)))
 		  (setq args (append args (list "-U" real-user)))
@@ -495,10 +494,11 @@ pass to the OPERATION."
 		    ;; Use an asynchronous processes.  By this,
 		    ;; password can be handled.
 		    (let* ((default-directory tmpdir)
-			   (p (start-process-shell-command
+			   (p (apply
+			       'start-process
 			       (tramp-get-connection-name v)
 			       (tramp-get-connection-buffer v)
-			       (mapconcat 'identity args " "))))
+			       tramp-smb-program args)))
 
 		      (tramp-message
 		       v 6 "%s" (mapconcat 'identity (process-command p) " "))
@@ -929,101 +929,109 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
   "Like `insert-directory' for Tramp files."
   (setq filename (expand-file-name filename))
   (unless switches (setq switches ""))
+  ;; Mark trailing "/".
+  (when (and (zerop (length (file-name-nondirectory filename)))
+	     (not full-directory-p))
+    (setq switches (concat switches "F")))
   (if full-directory-p
       ;; Called from `dired-add-entry'.
       (setq filename (file-name-as-directory filename))
     (setq filename (directory-file-name filename)))
   (with-parsed-tramp-file-name filename nil
-    (save-match-data
-      (let ((base (file-name-nondirectory filename))
-	    ;; We should not destroy the cache entry.
-	    (entries (copy-sequence
-		      (tramp-smb-get-file-entries
-		       (file-name-directory filename)))))
+    (with-tramp-progress-reporter v 0 (format "Opening directory %s" filename)
+      (save-match-data
+	(let ((base (file-name-nondirectory filename))
+	      ;; We should not destroy the cache entry.
+	      (entries (copy-sequence
+			(tramp-smb-get-file-entries
+			 (file-name-directory filename)))))
 
-	(when wildcard
-	  (string-match "\\." base)
-	  (setq base (replace-match "\\\\." nil nil base))
-	  (string-match "\\*" base)
-	  (setq base (replace-match ".*" nil nil base))
-	  (string-match "\\?" base)
-	  (setq base (replace-match ".?" nil nil base)))
+	  (when wildcard
+	    (string-match "\\." base)
+	    (setq base (replace-match "\\\\." nil nil base))
+	    (string-match "\\*" base)
+	    (setq base (replace-match ".*" nil nil base))
+	    (string-match "\\?" base)
+	    (setq base (replace-match ".?" nil nil base)))
 
-	;; Filter entries.
-	(setq entries
-	      (delq
-	       nil
-	       (if (or wildcard (zerop (length base)))
-		   ;; Check for matching entries.
-		   (mapcar
-		    (lambda (x)
-		      (when (string-match
-			     (format "^%s" base) (nth 0 x))
-			x))
-		    entries)
-		 ;; We just need the only and only entry FILENAME.
-		 (list (assoc base entries)))))
+	  ;; Filter entries.
+	  (setq entries
+		(delq
+		 nil
+		 (if (or wildcard (zerop (length base)))
+		     ;; Check for matching entries.
+		     (mapcar
+		      (lambda (x)
+			(when (string-match
+			       (format "^%s" base) (nth 0 x))
+			  x))
+		      entries)
+		   ;; We just need the only and only entry FILENAME.
+		   (list (assoc base entries)))))
 
-	;; Sort entries.
-	(setq entries
-	      (sort
-	       entries
-	       (lambda (x y)
-		 (if (string-match "t" switches)
-		     ;; Sort by date.
-		     (tramp-time-less-p (nth 3 y) (nth 3 x))
-		   ;; Sort by name.
-		   (string-lessp (nth 0 x) (nth 0 y))))))
+	  ;; Sort entries.
+	  (setq entries
+		(sort
+		 entries
+		 (lambda (x y)
+		   (if (string-match "t" switches)
+		       ;; Sort by date.
+		       (tramp-time-less-p (nth 3 y) (nth 3 x))
+		     ;; Sort by name.
+		     (string-lessp (nth 0 x) (nth 0 y))))))
 
-	;; Handle "-F" switch.
-	(when (string-match "F" switches)
+	  ;; Handle "-F" switch.
+	  (when (string-match "F" switches)
+	    (mapc
+	     (lambda (x)
+	       (when (not (zerop (length (car x))))
+		 (cond
+		  ((char-equal ?d (string-to-char (nth 1 x)))
+		   (setcar x (concat (car x) "/")))
+		  ((char-equal ?x (string-to-char (nth 1 x)))
+		   (setcar x (concat (car x) "*"))))))
+	     entries))
+
+	  ;; Print entries.
 	  (mapc
 	   (lambda (x)
-	     (when (not (zerop (length (car x))))
-	       (cond
-		((char-equal ?d (string-to-char (nth 1 x)))
-		 (setcar x (concat (car x) "/")))
-		((char-equal ?x (string-to-char (nth 1 x)))
-		 (setcar x (concat (car x) "*"))))))
-	   entries))
+	     (when (not (zerop (length (nth 0 x))))
+	       (when (string-match "l" switches)
+		 (let ((attr
+			(when (tramp-smb-get-stat-capability v)
+			  (ignore-errors
+			    (file-attributes filename 'string)))))
+		   (insert
+		    (format
+		     "%10s %3d %-8s %-8s %8s %s "
+		     (or (nth 8 attr) (nth 1 x)) ; mode
+		     (or (nth 1 attr) 1) ; inode
+		     (or (nth 2 attr) "nobody") ; uid
+		     (or (nth 3 attr) "nogroup") ; gid
+		     (or (nth 7 attr) (nth 2 x)) ; size
+		     (format-time-string
+		      (if (tramp-time-less-p
+			   (tramp-time-subtract (current-time) (nth 3 x))
+			   tramp-half-a-year)
+			  "%b %e %R"
+			"%b %e  %Y")
+		      (nth 3 x)))))) ; date
 
-	;; Print entries.
-	(mapc
-	 (lambda (x)
-	   (when (not (zerop (length (nth 0 x))))
-	     (let ((attr
-		    (when (tramp-smb-get-stat-capability v)
-		      (ignore-errors
-			(file-attributes filename 'string)))))
-	       (insert
-		(format
-		 "%10s %3d %-8s %-8s %8s %s "
-		 (or (nth 8 attr) (nth 1 x)) ; mode
-		 (or (nth 1 attr) 1) ; inode
-		 (or (nth 2 attr) "nobody") ; uid
-		 (or (nth 3 attr) "nogroup") ; gid
-		 (or (nth 7 attr) (nth 2 x)) ; size
-		 (format-time-string
-		  (if (tramp-time-less-p
-		       (tramp-time-subtract (current-time) (nth 3 x))
-		       tramp-half-a-year)
-		      "%b %e %R"
-		    "%b %e  %Y")
-		  (nth 3 x)))) ; date
 	       ;; We mark the file name.  The inserted name could be
-	       ;; from somewhere else, so we use the relative file
-	       ;; name of `default-directory'.
+	       ;; from somewhere else, so we use the relative file name
+	       ;; of `default-directory'.
 	       (let ((start (point)))
 		 (insert
 		  (format
 		   "%s\n"
 		   (file-relative-name
 		    (expand-file-name
-		     (nth 0 x) (file-name-directory filename)))))
+		     (nth 0 x) (file-name-directory filename))
+		    (when full-directory-p (file-name-directory filename)))))
 		 (put-text-property start (1- (point)) 'dired-filename t))
 	       (forward-line)
-	       (beginning-of-line))))
-	 entries)))))
+	       (beginning-of-line)))
+	   entries))))))
 
 (defun tramp-smb-handle-make-directory (dir &optional parents)
   "Like `make-directory' for Tramp files."
@@ -1270,6 +1278,8 @@ target of the symlink differ."
 
 	    ;; We must also flush the cache of the directory, because
 	    ;; `file-attributes' reads the values from there.
+	    (tramp-flush-file-property v1 (file-name-directory v1-localname))
+	    (tramp-flush-file-property v1 v1-localname)
 	    (tramp-flush-file-property v2 (file-name-directory v2-localname))
 	    (tramp-flush-file-property v2 v2-localname)
 	    (unless (tramp-smb-get-share v2)
@@ -1342,7 +1352,7 @@ target of the symlink differ."
 		;; Use an asynchronous processes.  By this, password can
 		;; be handled.
 		(let ((p (apply
-			  'start-process-shell-command
+			  'start-process
 			  (tramp-get-connection-name v)
 			  (tramp-get-connection-buffer v)
 			  tramp-smb-acl-program args)))
